import SCATPY
from pySCATMECH.model import *
from pySCATMECH.mueller import *
import numpy as np

class BRDF_Model(Model):
    """
    A class for handling BRDF Models. 

    Arguments *args and **kwargs can contain parameters passed to the model. 

    One form is to pass a dict of {"parameter" : value, ... }.

    The keyword argument 'wavelength' is automatically changed to 'lambda',
    since `lambda` is a Python keyword.  However, as a str, it should be 
    'lambda'.

    Parameters
    ----------
    name : str
        The name of a SCATMECH model inheriting BRDF_Model

    """
    
    def __init__(self,*args, **kwargs):

        self.newer = SCATPY.Get_BRDF_Model
        self.freer = SCATPY.Free_BRDF_Model
        if len(args)>0 and type(args[0])==str:
                self.handle = self.newer(args[0])
                args = args[1::]
        else:
            self.handle = self.newer("Microroughness_BRDF_Model")
        self.setParameters(*args, **kwargs)


    def __del__(self):
        SCATPY.Free_BRDF_Model(self.handle)

    def MuellerBRDF(self,
                    thetai=0, thetas=0, phis=0, rotation=0, coords="psps"):
        """
        Evaluate the Mueller matrix BRDF in a given geometry, and coordinate 
        system.

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians.
        thetas : float
            Polar angle of viewing in radians.
        phis: float, optional
            Azimuthal angle of viewing in radians.
            (Note: Specular occurs when thetai = thetas, phis = 0)
            (Default is 0)
        rotation : float, optional
            Sample rotation angle in radians. (Default is 0)
        coords : string, optional
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane." (Default is "psps")

        Returns
        -------
        BRDF : MuellerMatrix 
            The Mueller matrix BRDF
        """
        return MuellerMatrix(SCATPY.BRDF(self.handle,
                                         thetai, thetas, phis,
                                         rotation, coords))

    def JonesBRDF(self,
                  thetai=0, thetas=0, phis=0, rotation=0, coords="psps"):
        """
        Evaluate the Jones matrix BRDF in a given geometry, and coordinate 
        system. The result, when typecast to a Mueller matrix, is the Mueller
        matrix BRDF, but maintains phase, if the model has phase information.

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians.
        thetas : float
            Polar angle of viewing in radians.
        phis: float, optional
            Azimuthal angle of viewing in radians.
            (Note: Specular occurs when thetai = thetas, phis = 0)
            (Default is 0)
        rotation : float, optional
            Sample rotation angle in radians. (Default is 0)
        coords : string, optional
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane." (Default is "psps")

        Returns
        -------
        BRDF : numpy.array of complex (2x2)
            The Jones matrix BRDF
        """
        return np.array(SCATPY.BRDFJones(self.handle,
                                         thetai, thetas, phis,
                                         rotation, coords))

    def BRDF(self,
             thetai=0, thetas=0, phis =0, rotation=0,
             coords="psps",
             inc=StokesVector([1,0,0,0]),
             sens=StokesVector([1,0,0,0])):
        """
        Return the scalar BRDF in a given geometry.

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians.
        thetas : float
            Polar angle of viewing in radians. 
        phis : float, optional
            Azimuthal angle of viewing in radians.
            (Note: Specular occurs when thetai = thetas, phis = 0)
            (default is 0)
        rotation : float, optional
            Sample rotation angle in radians. (Default is 0)
        coords : string, optional
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane". (Default is `psps`)
        inc : StokesVector or 4-list of float or int, optional
            Incident polarization as Stokes vector.
            (Default is unpolarized)
        sens : StokesVector or 4-list of float or int, optional
            Polarization sensitivity of the viewer as a Stokes vector. 
            (Default is unpolarized)

        Returns
        -------
        BRDF : float
            The Mueller matrix BRDF for the given incident and viewing 
            directions and polarizations.
        """
        return float((StokesVector(sens) @
                      MuellerMatrix(SCATPY.BRDF(self.handle, thetai, thetas,
                                                phis, rotation, coords))) @
                     StokesVector(inc))




import SCATPY

class Model:
    """
    Base class for Models.
    """

    def __init__(self,*args,**kwargs):
        self.newer = SCATPY.Get_Model
        self.freer = SCATPY.Free_Model

        if len(args)>0 and type(args[0])==str:
                self.handle = self.newer(args[0])
                args = args[1::]
        else:
            self.handle = self.newer("Microroughness_BRDF_Model")
        self.setParameters(*args, **kwargs)

    def setParameters(self,*args, **kwargs):
        """
        Sets one or more parameters for the model. 

        Parameters
        ----------
        parameter_values : dict 
                           Argument is a dictionary with with str keys being 
                           the parameters and values being the values. 
                           `None` key defines model name.

        **kwargs : 
                    parameters can be set by parameter name, 
                    `lambda` should be replace with `wavelength`, since
                    lambda is a Python keyword.

        Returns
        -------
        Model


        """
        # Change `wavelength` to `lambda` in kwargs...
        if "wavelength" in kwargs:
            kwargs["lambda"]=kwargs["wavelength"]
            del kwargs["wavelength"]

        parameter_values = dict()
        for arg in args:
            if hasattr(arg,"getParameters"):
                arg = arg.getParameters()
            parameter_values.update(arg)
            
        parameter_values.update(kwargs)
            
        for parameter,value in parameter_values.items():
            
            if hasattr(value,"getModelDict"):
                # value is a FilmStack...
                value = value.getModelDict()
            elif hasattr(value,"nkstr"):
                # value is a OpticalFunction...
                value = value.nkstr()
            elif type(value) is complex:
                # value is complex and needs to be converted to a str...
                value = "(%16g,%16g)" % (value.real,value.imag)

            if parameter==None:
                # Set the Model...
                self.freer(self.handle)
                self.handle = self.newer(value)
            elif type(value)==dict:
                dictnew = {}
                for subparameter,subvalue in value.items():
                    if hasattr(subvalue,'imag'):
                        if subvalue.imag==0:
                            subvalue = "%.16g" %subvalue.real
                        else:
                            subvalue = "(%.16g,%.16g)" % (subvalue.real,subvalue.imag)
                        
                    if type(subparameter)==type(None):
                        SCATPY.SetParameter(self.handle,
                                            str(parameter),
                                            str(subvalue))
                    else:
                        dictnew[parameter + "." + subparameter] = subvalue
                self.setParameters(dictnew)
            elif isinstance(value,Model):
                self.setParameters({str(parameter) : value.getParameters()})
            else:
                SCATPY.SetParameter(self.handle,str(parameter),str(value))
        return self

    def setParameter(self, parameter, value):
        """
        Sets a single parameter for the model.

        Arguments
        ---------
        parameter: str
                   Parameter name as a string.  
                   Subparameters must use `parameter.subparameter` notation.
        value: float, str, or complex
               Parameter value
        """
        return self.setParameters({parameter : value})
        
        if hasattr(value,'imag'):
            if value.imag==0:
                value = "%.16g" % value.real
            else:
                value = "(%.16g,%.16g)" % (value.real,value.imag)

        SCATPY.SetParameter(self.handle,str(parameter),value)
        return self
            
    def getParameter(self, parameter):
        """
        Gets a string corresponding to the value of a model parameter.
        Arguments:
        parameter: parameter name as a string
        Returns: parameter value
        """
        return SCATPY.GetParameter(self.handle, str(parameter))

    def getParameters(self):
        """
        Gets all of the parameters for the model.

        Returns
        -------
             dict with str keys
        """
        dict = SCATPY.GetParameterDictionary(self.handle)
        newdict = {None : self.getModelName()}
        for d in dict:
            newdict[d]=dict[d]['value']
        return newdict
        
    def printParameters(self):
        """
        Returns a printable string that includes all the parameters, 
        their values, and their descriptions.
        Recommend surround with print()
        """
        return SCATPY.PrintParameters(self.handle)

    def getParameterDictionary(self):
        """
        Returns a dictionary of dictionaries for each parameter that includes 
        values, descriptions, and types.
        """
        return SCATPY.GetParameterDictionary(self.handle)

    def askParameters(self):
        """
        Queries the user for parameter values.
        TODO FIX!
        """ 
        params = self.getParameterDictionary()
        for param in params.copy():
            if '.' in param:
                del params[param]
        del param
        result = {}
        for param,d in params.items():
            if d['type'][-4:]!="_Ptr":
                response = input(d['description']+" <"+ d['value'] + ">:")
                if response == '':
                    result[param] = str(d['value'])
                else:
                    result[param] = response
            else:
                result[param]= chooseModelAndParameters(d['type'][:-4])
        self.setParameters(result)

    def getModelName(self):
        """
        Returns the model's name
        """
        return SCATPY.GetModelName(self.handle)

    def __repr__(self):
        return self.getModelName()
    
    def __str__(self):
        """
        Returns a string version of the model, which is a pretty-printed map 
        of maps.
        """
        result = {None : self.getModelName()}
        params = self.getParameters()
        for p in params:
            result[p] = params[p]
        return strDictOfDict(createDictOfDict(result))
        
    def __del__(self):
        try:
            SCATPY.Free_Model(self.handle)
        except:
            pass
                
def getModelDictionary(model="Model"):
    """
    Returns a dictionary containing all the models inheriting model and their 
    descriptions.
    """
    return SCATPY.GetModelDictionary(model)

def chooseModel(model):
    """
    Presents user with a choice of models and returns the user's choice, by 
    name.
    """
    choices = getModelDictionary(model)
    i = 0
    for m in choices:
        choices[m][None] = i
        i=i+1
        
    valid = False
    while not valid:
        i = 0
        for m,d in choices.items():
            print(d[None]," : ", m," : ",d['description'])
            i = i+1
        response = input("Choose one:")
        if response in choices:
            return response
        for m,d in choices.items():
            try:
                if d[None]==int(response):
                    return m
            except ValueError:
                pass
            
def chooseModelAndParameters(model):
    """
    Presents user with a choice of models and asks for parameters.
    """
    modelname = chooseModel(model)
    result = {None:modelname}
    model = Model(modelname)
    params = model.getParameterDictionary()
    del model
    for param in params.copy():
        if '.' in param:
            del params[param]
    for param,d in params.items():
        if d['type'][-4:]!="_Ptr":
            response = input(d['description']+" <"+ d['value'] + ">:")
            if response == '':
                result[param] = str(d['value'])
            else:
                result[param] = response
        else:
             result[param]= chooseModelandParameters(d['type'][:-4])
    return result

def createDictOfDict(theDict):
    """
    Creates a dictionary of dictionaries for a simple dictionary
    with dot-delimited keys. The result[gparent][parent][child]
    theDict["gparent.parent.child"]-->result["gparent"]["parent"]["child"]
    theDict["gparent.parent"]--> result[gparent][parent][None]
    """
    result = {}
    for p in theDict:
        if p==None:
            result[p] = theDict[p] 
        elif '.' in p:
            parent = p[:p.find('.')]
            child = p[p.find('.')+1:]
            if type(result[parent])!=dict:
                result[parent] = {None : result[parent]}
            result[parent][child] = theDict[p] 
        else:
            result[p] = theDict[p]
    for p in result:
        if type(result[p])==dict:
            result[p] = createDictOfDict(result[p])
    return result

def printDictOfDict(thedict):
    print(strDictOfDict(thedict))
    
def strDictOfDict(thedict,indent="",result=None):
    """
    Prints a dictionary of dictionaries in a readable format
    """
    if result==None:
        result = "{"
    else:
        result += "{"
    i=0
    for a in thedict:
        if i:
            result += ",\n"
        b = "'" + a + "'" if a!=None else "None"
        if type(thedict[a])==dict:
            if i:
                result += indent + b + " : "
            else:
                result += b + " : "
            result = strDictOfDict(thedict[a],indent+(len(a)+7)*' ',result)
        else:
            c = "'"+thedict[a]+"'"
            if i:
                result += indent + b + " : " + c
            else:
                result += b + " : " + c
        i = i+1
    
    result += "}"
    return result



def DialogGetModel(top):
    """
    Opens a dialog box with a selection of Models inheriting top and 
    returns the choice.

    Parameters
    ----------
    top : str
          The top level class from which to choose a child class

    Returns
    -------
    str : The name of the selected child class

    Requires
    --------
    tkinter
    """
    import tkinter
    from tkinter import ttk
    from tkinter import simpledialog

    class ModelDialog:
        # Opens a dialog box containing a bunch of choices. User 
        # double-clicks or presses enter to choose.
        # On completion, result contains the result.
        def ok(self,event):
            parameter = self.tr.selection()[0]
            self.result = parameter
            self.top.destroy()

        def __init__(self, title, choices):

            self.choices=choices

            top = self.top = tkinter.Tk()
            top.title(title)

            self.tr = tkinter.ttk.Treeview(top)
            self.tr["columns"] = ("Description")
            for t in self.tr["columns"]:
                self.tr.column(t)
                self.tr.heading(t,text=t)
                rows = 0
                cols = 0
            for c in self.choices:
                self.tr.insert("", "end", c, text=c,
                               values = (self.choices[c]["description"],))
                cols = max(cols,len(c)+len(self.choices[c]["description"]))
                rows += 1
                self.tr.pack(expand=True, fill="both")
                self.tr.bind('<Double-1>', self.ok)
                self.tr.bind('<Return>', self.ok)

            self.result = None
            top.geometry("%dx%d" % (cols*8,rows*22))
            top.wait_window(self.top)

    dialog = ModelDialog("Select model:", getModelDictionary(top))
    try:
        model = Model(dialog.result)
    except:
        model = DialogGetModel(top)

    return model
        

def DialogGetModelParameters(model):
    """
    Opens a dialog box with a selection of Models inheriting top and 
    returns the choice.

    Parameters
    ----------
    model : Model
            The model to fill parameters out with.
    
    Returns
    -------
    Model
    The model with new parameters

    Requires
    --------
    tkinter
    """
    import tkinter
    from tkinter import ttk
    from tkinter import simpledialog

    class ModelParameterDialog():
        # Creates a Treeview with all of a model's parameters. User can change 
        # values as desired. Upon completion, the model is set as desired.
        def __init__(self, model):
            self.model = model
            top = self.top = tkinter.Tk()
            self.top.title("Model Parameters")
            self.tree = tkinter.ttk.Treeview(self.top)
            self.tree["columns"] = ("value","description","type")
            self.tree.column("value")
            self.tree.column("description")
            self.tree.column("type")
            self.tree.heading("value", text="Value")
            self.tree.heading("description", text="Description")
            self.tree.heading("type", text="Type")
            self.reload()
            self.tree.pack(expand=True,fill='both')
            top.wait_window(self.top)

        def reload(self):
            x = self.tree.get_children()
            for t in x:
                self.tree.delete(t)
            params = self.model.getParameterDictionary()
            for p in params:
                if "_Ptr" in params[p]['type']:
                    if '.' in p:
                        parent = p[:p.rfind('.')]
                        child = p[p.rfind('.')+1:]
                    else:
                        parent = ""
                        child = p
                elif '.' in p:
                    parent = p[:p.rfind('.')]
                    child = p[p.rfind('.')+1:]
                else:
                    parent = ""
                    child = p

                self.tree.insert(parent,"end",p,text=child,
                                 values=(str(params[p]["value"]),
                                         params[p]["description"],
                                         params[p]["type"]))

            self.tree.bind('<Double-1>',self.clickon)
            self.tree.bind('<Return>',self.clickon)

        def clickon(self,event):
            parameter = self.tree.selection()[0]
            v = self.tree.item(parameter)['values']
            prompt = ("Parameter = " + parameter +
                      "\n" + str(v[1]) +
                      "\nDatatype = " + str(v[2]) +
                      "\nCurrent value = " + str(v[0]) +
                      "\nEnter new value:")
            if v[2][-4:]=='_Ptr':
                dialogresult = DialogGetModel(v[2][:-4])
                if (dialogresult!=None):
                    self.model.setParameters({parameter : dialogresult})
                    self.reload()
            else:
                answer = tkinter.simpledialog.askstring("Input", prompt,
                                                        parent=self.tree)
                if answer!="" and answer!=None:
                    temp = self.tree.item(parameter)['values']
                    temp[0]=answer
                    self.tree.item(parameter,values=temp)
                    self.model.setParameter(parameter,answer)
                    
    return ModelParameterDialog(model).model
  



from pySCATMECH.brdf import *
from pySCATMECH.local import *
import matplotlib.pyplot as plt
import numpy as np

class SolidAngle:
    """
    Abstract class for a solid angle on a hemisphere.

    Inherited classes are expected to define inside(self,v) where
    v contains directional cosines and returns a tuple of 
    bool (`True` if (x,y) is inside solid angle, `False` otherwise) and
    the Stokes vector sensitivity.

    The parent class provides __and__, __or__, and __invert__
    functionality, so that solid angles can be defined by combining
    other solid angle classes.
    """
    def __and__(self, other):
        """
        Combines two SolidAngle with an `and` operation.
        Returns a new SolidAngle.
        """
        class newSolidAngle(SolidAngle):
            def __init__(s, a, b):
                s.a = a
                s.b = b

            def inside(s, v):
                ai = s.a.inside(v)
                bi = s.b.inside(v)
                if ai[0] and not bi[0]:
                    return False, ai[1]
                elif not ai[0] and bi[0]:
                    return False, bi[1]
                elif not ai[0] and not bi[0]:
                    return False, (ai[1] + bi[1])/2
                else:
                    return True, (ai[1] + bi[1])/2
        return newSolidAngle(self, other)
    
    def __or__(self, other):
        """
        Combines two SolidAngle with an `or` operation.
        Returns a new SolidAngle.
        """
        class newSolidAngle(SolidAngle):
            def __init__(s, a, b):
                s.a = a
                s.b = b
            def inside(s, v):
                ai = s.a.inside(v)
                bi = s.b.inside(v)
                if ai[0] and not bi[0]:
                    return True, ai[1]
                elif not ai[0] and bi[0]:
                    return True, bi[1]
                elif not ai[0] and not bi[0]:
                    return False, (ai[1] + bi[1])/2
                else: 
                    return True, (ai[1] + bi[1])/2
        return newSolidAngle(self, other)

    def __invert__(self):
        """
        Inverts a SolidAngle.
        Returns a new SolidAngle.
        """
        class newSolidAngle(SolidAngle):
            def __init__(s, a):
                s.a = a
            def inside(s, v):
                ai = s.a.inside(v)
                return not ai[0], ai[1] 
        return newSolidAngle(self)

class Hemisphere(SolidAngle):
    """
    A solid angle defined by the entire hemisphere.
    """
    def __init__(self, sensitivity = StokesVector(1,0,0,0), polphi = None):
        """
        Parameters
        ----------
        
        sensitivity : StokesVector, optional
                      The Stokes vector sensitivity, where signal is assumed
                      to be the inner product of this value with the incident
                      Stokes vector. The Stokes vector is represented in an s-p
                      basis set.
                      (default: No polarization sensitivity)

        polphi : None or float, optional
                 The azimuthal angle in radians where the polarization
                 sensitivity is defined. If None (default), the polarization 
                 sensitivity is defined in an x-y basis set.
        """
        if polphi is None:
            self.sensitivity = sensitivity
        else:
            self.sensitivity = MuellerRotator(polphi) @ sensitivity
        
    def inside(self, v):
        """
        Parameters
        ----------

        v : tuple of float
            Direction represented as directional cosines

        Returns
        -------
        is_inside : bool
                    True if v is inside solid angle, False otherwise

        sensitivity : StokesVector
                      The Stokes vector sensitivity
        """
        v2 = v[0]**2 + v[1]**2
        if v2>=1: return False, self.sensitivity
        return True, self.sensitivity
    
class ProjectedPolygon(SolidAngle):
    """
    A solid angle defined by an array of (x,y) points in projected cosine space.
    """
    def __init__(self, boundary,
                 sensitivity = StokesVector(1,0,0,0),
                 polphi = None):
        """
        Parameters
        ----------
        
        sensitivity : StokesVector, optional
                      The Stokes vector sensitivity, where signal is assumed
                      to be the inner product of this value with the incident
                      Stokes vector. 
                      (default: No polarization sensitivity)

        polphi : None, float, or str, optional
                 If None or "sp" (default), the polarization sensitivity is 
                 defined in an s-p basis at the centroid of the projected 
                 solid angle.
                 If float, the azimuthal angle in radians where the polarization
                 sensitivity is defined in the s-p basis.
                 If "xy", the sensitivity is defined in an x-y basis.
        """
        self.boundary = boundary
        self.boundary.append(boundary[0])

        # Find centroid:
        area = 0
        centroidx = 0
        centroidy = 0
        for a,b in zip(boundary,boundary[1:]+boundary[:1]):
            temp = a[0]*b[1]-a[1]*b[0]
            area += temp
            centroidx += (a[0]+b[0])*temp
            centroidy += (a[1]+b[1])*temp
        centroidx /= 6*area
        centroidy /= 6*area

        centphi = math.atan2(centroidy,centroidx)
        
        if polphi is None or polphi == 'sp':
            self.sensitivity = MuellerRotator(-centphi) @ sensitivity
        elif polphi=='xy':
            self.sensitivity = sensitivity
        else:
            self.sensitivity = MuellerRotator(-polphi) @ sensitivity

            
    def inside(self, v):
        """
        Parameters
        ----------

        v : tuple of float
            Direction represented as directional cosines

        Returns
        -------
        is_inside : bool
                    True if v is inside solid angle, False otherwise

        sensitivity : StokesVector
                      The Stokes vector sensitivity
        """
        v2 = v[0]**2 + v[1]**2        
        if v2>1:
            return False, self.sensitivity
        count = 0
        b0 = self.boundary[0]
        x, y = v[0], v[1]
        for b1 in self.boundary[1:]:
            x0, x1, y0, y1 = b0[0], b1[0], b0[1], b1[1]
            b0 = b1
            if x0 < x and x1 <= x: pass
            elif x0 > x and x1 >= x: pass
            elif y0 < y and y1 <= y: pass
            elif y0 >= y and y1 > y:
                if (x0 >= x and x1 < x) or (x0 <= x and x1 > x):
                    count = count + 1
            elif x1 == x0:
                if (y0 >= x and y1 < y) or ( y0 <= x and y1 > y):
                    count = count + 1
            else:
                yc = y0 + (y1 - y0) * (x - x0) / (x1 - x0)
                if yc > y:
                    count = count + 1
        return count % 2 == 1, self.sensitivity
        
class EllipticalCone(SolidAngle):
    """
    A solid angle defined as an elliptical circular cone.
    """
    def __init__(self, theta = 0, phi = 0, alpha = 0, gamma = 0,
                 sensitivity = StokesVector(1,0,0,0), polphi = None):
        """
        A solid angle defined as an elliptical circular cone.

        Parameters
        ----------
        
        theta : float, optional
                Polar angle of center of cone [rad]
                (Default is 0)

        phi : float, optional
              Azimuth angle of center of cone [rad]
              (Default is 0)

        alpha : float or tuple of float
                If a float, half angle of cone [rad]
                If a tuple of float, (alphax, alphay), where
                alphax is the half angle of cone along the x-z plane before 
                rotation [rad], and alphay is the half angle of cone along the 
                y-z plane before rotation [rad]

        gamma : float, optional
                Rotation about the z axis [rad]
                (Default is 0)

        sensitivity : StokesVector, optional
                      The Stokes vector sensitivity, where signal is assumed
                      to be the inner product of this value with the incident
                      Stokes vector. The Stokes vector is represented in an s-p
                      basis set.
                      (default: No polarization sensitivity)

        polphi : None, float, or str, optional
                 If None or "sp" (default), the polarization sensitivity is 
                 defined in an s-p basis at the centroid of the projected 
                 solid angle.
                 If float, the azimuthal angle in radians where the polarization
                 sensitivity is defined in the s-p basis.
                 If "xy", the sensitivity is defined in an x-y basis.

        """
        if hasattr(alpha, '__len__'):
            alphax = alpha[0]
            alphay = alpha[1]
        else:
            alphax = alphay = alpha

        if polphi is None or polphi == 'sp':
            self.sensitivity = MuellerRotator(-phi) @ sensitivity
        elif polphi == 'xy':
            self.sensitivity = sensitivity
        else:
            self.sensitivity = MuellerRotator(-polphi) @ sensitivity

        sin = np.sin
        cos = np.cos
        
        self.theta = theta
        self.phi = phi
        self.alphax = alphax
        self.alphay = alphay
        self.gamma = gamma
        v = np.array([sin(theta) * cos(phi), sin(theta) * sin(phi), cos(theta)])
        s = perpto(v, (1, 0, 0))
        p = perpto(v, s)
        self.rotation_matrix = np.array([p, s, v])
        self.sinalphax = sin(alphax)
        self.sinalphay = sin(alphay)
        v = np.array([cos(gamma), sin(-gamma), 0])
        s = perpto(v, (0, 0, 1))
        p = perpto(v, s)
        self.rotation_matrix = np.array([v, s, p]) @ self.rotation_matrix
        
    def inside(self, v):
        """
        Parameters
        ----------

        v : tuple of float
            Direction represented as directional cosines

        Returns
        -------
        is_inside : bool
                    True if v is inside solid angle, False otherwise

        sensitivity : StokesVector
                      The Stokes vector sensitivity
        """
        v2 = v[0]**2 + v[1]**2
        if v2 > 1: return False, self.sensitivity
        v = np.array([v[0], v[1], np.sqrt(1 - v2)])
        w = self.rotation_matrix @ v
        if w[0]**2 / self.sinalphax**2 + w[1]**2 / self.sinalphay**2 < 1:
            return True, self.sensitivity
        return False, self.sensitivity

# A pseudonym 
CircularCone = EllipticalCone

class RectangularCone(SolidAngle):
    """
    A solid angle defined by a rectangular collection cone.
    """
    def __init__(self, theta = 0, phi = 0, alpha = 0, gamma = 0,
                 sensitivity = StokesVector(1,0,0,0), polphi = None):
        """
        A solid angle defined by a rectangular collection cone.

        Parameters
        ----------
        
        theta : float, optional
                Polar angle of center of cone [rad]
                (Default is 0)

        phi : float, optional
              Azimuth angle of center of cone [rad]
              (Default is 0)

        alpha : float or tuple of float
                If a float, half angle of cone [rad]
                If a tuple of float, (alphax, alphay), where
                alphax is the half angle of cone along the x-z plane before 
                rotation [rad], and alphay is the half angle of cone along the 
                y-z plane before rotation [rad]

        gamma : float, optional
                Rotation about the z axis [rad]
                (Default is 0)
        
        sensitivity : StokesVector, optional
                      The Stokes vector sensitivity, where signal is assumed
                      to be the inner product of this value with the incident
                      Stokes vector. The Stokes vector is represented in an s-p
                      basis set.
                      (default: No polarization sensitivity)

        polphi : None, float, or str, optional
                 If None or "sp" (default), the polarization sensitivity is 
                 defined in an s-p basis at the centroid of the projected 
                 solid angle.
                 If float, the azimuthal angle in radians where the polarization
                 sensitivity is defined in the s-p basis.
                 If "xy", the sensitivity is defined in an x-y basis.
        """
        if hasattr(alpha, '__len__'):
            alphax = alpha[0]
            alphay = alpha[1]
        else:
            alphax = alphay = alpha

        if polphi is None or polphi == 'sp':
            self.sensitivity = MuellerRotator(-phi) @ sensitivity
        elif polphi == 'xy':
            self.sensitivity = sensitivity
        else:
            self.sensitivity = MuellerRotator(-polphi) @ sensitivity

        sin = np.sin
        cos = np.cos

        self.theta = theta
        self.phi = phi
        self.alphax = alphax
        self.alphay = alphay
        v = np.array([sin(theta) * cos(phi), sin(theta) * sin(phi), cos(theta)])
        s = perpto(v, (1, 0, 0))
        p = perpto(v, s)
        self.rotation_matrix = np.array([p, s, v])
        self.sinalphax = sin(alphax)
        self.sinalphay = sin(alphay)
        v = np.array([cos(gamma), -sin(gamma), 0])
        s = perpto(v, (0, 0, 1))
        p = perpto(v, s)
        self.rotation_matrix = np.array([v, s, p]) @ self.rotation_matrix
    
    def inside(self, v):
        """
        Parameters
        ----------

        v : tuple of float
            Direction represented as directional cosines

        Returns
        -------
        is_inside : bool
                    True if v is inside solid angle, False otherwise

        sensitivity : StokesVector
                      The Stokes vector sensitivity
        """
        v2 = v[0]**2 + v[1]**2
        if v2 > 1: return False, self.sensitivity
        
        v = np.array([v[0], v[1], np.sqrt(1-v2)])
        w = self.rotation_matrix @ v
        if abs(w[0]) < self.sinalphax and abs(w[1]) < self.sinalphay:
            return True, self.sensitivity
        return False, self.sensitivity

# A pseudonym 
SquareCone = RectangularCone

class Integrator():
    """Class designed to integrate scattering models.

    Class designed to integrate a BRDF_Model (to obtain reflectance) or
    a Local_BRDF_Model (to obtain scattering cross section). 

 
    Attributes
    ----------
    x, y, z : ndarray(float)
           The x and y projected cosines

    w    : ndarray(float)
           The weights

    sens : list of StokesVectors
           The Stokes sensitivities

    theta : ndarray(float)
            The polar angles in radians.

    phi : ndarray(float)
          The azimuthal angles in radians.
 
    """    
    def __init__(self, gridsize, detector, type=1):
        """
        Parameters
        ----------
    
        gridsize : float 
               The size of the integration step in radians. In the direction 
               of the surface normal, the step size is asin(gridsize). 
        
        detector : SolidAngle
               A description of the detection system.

        type : int
           The method used for sampling type
           1 for evenly spaced integration in angle space, 
             [thetas*cos(phis),thetas*sin(phis)]
           2 for evenly spaced integration in solid angle, 
             [sqrt(1-cos(thetas))*cos(phis),sqrt(1-cos(thetas))*sin(phis)]
           3 for evenly spaced integration in projected solid angle, 
             [sin(thetas)*cos(phis),sin(thetas)*sin(phis)]
        """
        if (type == 1):
            def sinc(t):
                if t!=0: return math.sin(t)/t
                return 1
    
            def trans(x, y):
                theta = math.sqrt(x**2 + y**2) * pi / 2
                if theta > pi/2: return (1, 1, 0)
                sintheta = math.sin(theta)
                phi = math.atan2(y, x)
                return (sintheta * math.cos(phi),
                        sintheta * math.sin(phi),
                        math.cos(theta) * sinc(theta) * (gridsize*pi/2)**2)
        
            temp = [trans(x, y)
                    for x in np.arange(-1, 1, gridsize)
                    for y in np.arange(-1, 1, gridsize)
                    if x**2 + y**2 < 1]

        if (type == 2):
        
            def trans(x, y):
                sqrt2 = math.sqrt(2)
                costheta = 1 - x**2 - y**2
                sintheta = math.sqrt(1 - costheta**2)
                phi = math.atan2(y, x)
                return (sintheta * math.cos(phi), 
                        sintheta * math.sin(phi),
                        2 * costheta * gridsize**2)
        
            temp = [trans(x, y)
                    for x in np.arange(-1, 1, gridsize)
                    for y in np.arange(-1, 1, gridsize)
                    if x**2 + y**2 < 1]

        if (type==3):

            temp = [(x, y, gridsize**2)
                    for x in np.arange(-1, 1, gridsize)
                    for y in np.arange(-1, 1, gridsize)
                    if x**2 + y**2 < 1]

        dtemp = [(x, y, w, detector.inside((x, y))) for x, y, w in temp]
        dtemp = [(x, y, w, s[1]) for x, y, w, s in dtemp if s[0]]
                 
        self.x = np.array([t[0] for t in dtemp])
        self.y = np.array([t[1] for t in dtemp])
        self.z = np.sqrt(1 - self.x**2 - self.y**2)
        self.theta = np.arccos(self.z)
        self.phi = np.arctan2(self.y,self.x)
        self.w = np.array([t[2] for t in dtemp])
        self.sens = [t[3] for t in dtemp]
    
    def PlotSamplingPoints(self, figsize=5,
             expand=1, color='b', circlecolor='r'):
        """
        Graphs the integration points on the projected hemisphere.
    
        Parameters
        ----------
    
        figsize : float, optional
              Size of figure, which will be square (default is 5)
              
        expand : float, optional
             Scaling of sampling point size (default is 1) 
    
        color : str, optional
            Color to use for the sampling points (default is 'b')
            
        circlecolor : str, optional
                  Color to use for the horizon (default is 'r') 
          
        Requires: matplotlib 
        """
        import matplotlib.pyplot as plt
        
        # Create a curve to show the horizon...
        thetas = np.linspace(0, 2*pi, 100)
        circlex = np.cos(thetas)
        circley = np.sin(thetas)
    
        plt.figure(figsize = (figsize, figsize))
        ax = plt.subplot(111)
        ax.scatter(self.x, self.y,
                   s = self.w * 300 * figsize**2 * expand, c=color) 
        ax.plot(circlex, circley, c = circlecolor)
        plt.show()
    
    def Reflectance(self, model, thetai, 
                    incpol = StokesVector(1,0,0,0),
                    rotation = 0):
        """
        Integrates a BRDF_Model and returns reflectance.
        
        Parameters
        ----------

        model :  BRDF_Model 
                 The model to be integrated.

        thetai : float
                 The incident angle in radians

        incpol : StokesVector
                 The incident Stokes vector

        rotation: float
                  The angle of rotation of the material in radians

        Returns
        -------

        reflectance : float
                      The integrated reflectance
        """
        weights = self.w
        return self.BRDF(model, thetai,
                         incpol=incpol, rotation=rotation) @ weights

    def BRDF(self, model, thetai, incpol=StokesVector(1,0,0,0), rotation=0):
        """
        Returns the BRDF at each integration point.

        Parameters
        ----------

        model : BRDF_Model
                The BRDF model to be integrated

        thetai : float
                The incident angle in radians

        incpol : StokesVector
                 The incident Stokes vector

        rotation: float
                  The angle of rotation of the material in radians

        Returns
        -------

        BRDF : np.array 
               An array of BRDF
        """
        return np.array([model.BRDF(thetai, theta, phi,
                                    rotation=rotation,
                                    coords="xyxy", inc=incpol,
                                    sens=sens)
                         for theta,phi,sens in zip(self.theta,
                                                   self.phi,
                                                   self.sens)])

    def CrossSection(self, model, thetai,
                     incpol = StokesVector(1,0,0,0),
                     rotation = 0):
        """
        Integrates a Local_BRDF_Model to obtain differential scattering 
        cross section.

        Parameters
        ----------

        model:  Local_BRDF_Model 
                The model to be integrated.

        thetai: float
                The incident angle in radians

        incpol : StokesVector
                 Incident Stokes vector

        Returns
        -------

        crosssection : float
                       The integrated cross section
        """
        weights = self.w / self.z
        return self.DSC(model, thetai, incpol) @ weights

    def DSC(self, model, thetai, incpol = StokesVector(1,0,0,0), rotation = 0):
        """
        Returns the differential scattering cross section at each 
        integration point.

        Parameters
        ----------

        model : Local_BRDF_Model
                The BRDF model to be integrated

        thetai: float
                The incident angle in radians

        incpol : StokesVector
                 Incident Stokes vector

        rotation: float
                  The angle of rotation of the material in radians

        Returns
        -------

        DSC : np.array 
              An array of differential scattering cross-section.
        """
        return np.array([model.DSC(thetai, theta, phi,
                                   coords="xyxy", rotation=rotation,
                                   inc=incpol, sens=sens)
                         for theta, phi, sens in zip(self.theta,
                                                     self.phi,
                                                     self.sens)])



import SCATPY
from pySCATMECH.model import *
from pySCATMECH.mueller import *

class Free_Space_Scatterer(Model):
    """A class for handling free space scatterer models"""

    def __init__(self, *args, **kwargs):
        """
        Creates and allocates a SCATMECH Free_Space_Scatterer 
        
        Parameters
        ----------
        name : str
            The name of a SCATMECH model inheriting Free_Space_Scatterer 

        args and kwargs : dict of str : str
            Initializing parameters for the model.  The dict items can be
            of the form 'param' : 'value', 'param.subparam' : 'value', 
            or 'param' : dict, where the first dict item is 
            None : 'submodelname' and 
            subsequent items are parameters of submodelname.
        
        Returns
        -------
        r : FSScatterer
            The object containing a handle to the requested Free_Space_Scatterer
        """

        self.newer = SCATPY.Get_Free_Space_Scatterer
        self.freer = SCATPY.Free_Free_Space_Scatterer
        if len(args)>0 and type(args[0])==str:
                self.handle = self.newer(args[0])
                args = args[1::]
        else:
            self.handle = self.newer("MieScatterer")
        self.setParameters(*args, **kwargs)

    def __del__(self):
        SCATPY.Free_Free_Space_Scatterer(self.handle)

    def DifferentialScatteringCrossSection(self, vi, vo):
        """
        Returns the Mueller matrix differential scattering cross section for 
        a given geometry

        Arguments
        ---------
        vi : 3-vector 
             The incident direction
        
        vo : 3-vector 
             The scattering direction

        Returns
        -------
        DSC : MuellerMatrix
              Mueller matrix differential scattering cross section
        """
        return JonesMueller(SCATPY.FSSjones(self.handle,
                                        vi[0], vi[1], vi[2],
                                        vo[0], vo[1], vo[2]))

    def ScatteringMatrix(self, vi, vo):
        """
        Returns the Mueller matrix differential scattering cross section for 
        a given geometry

        Arguments
        ---------
        vi : 3-vector 
             The incident direction
        
        vo : 3-vector 
             The scattering direction

        Returns
        -------
        s : numpy.array
              Scattering matrix as a Jones matrix
        """
        return np.array(SCATPY.FSSjones(self.handle,
                               vi[0], vi[1], vi[2],
                               vo[0], vo[1], vo[2]))

    def Extinction(self, v):
        """
        Returns the Mueller matrix extinction cross section for 
        a given geometry

        Arguments
        ---------
        v : 3-vector 
             The incident direction
        
        Returns
        -------
        e : MuellerMatrix
            Mueller matrix extinction cross section
        """
        return MuellerMatrix(SCATPY.FSSext(self.handle,v[0],v[1],v[2]))



import SCATPY
from pySCATMECH.model import *
from pySCATMECH.mueller import *


class CrossRCW_Model(Model):
    """
    Class handling rigorous coupled wave (RCW) analysis for cross gratings.
    """
    def __init__(self, *args, **kwargs):
        """
        Arguments and keyword arguments are passed to Model.setParameters
        and contain parameters and values
        """
        self.handle = SCATPY.Get_CrossRCW_Model()
        self.setParameters(*args, **kwargs)
        self.newer =  SCATPY.Get_CrossRCW_Model
        self.freer =  SCATPY.Free_CrossRCW_Model
        
    def __del__(self):
        SCATPY.Free_CrossRCW_Model(self.handle)

    def DiffractionEfficiency(self, i, j):
        """
        Returns the Mueller matrix diffraction efficiency for the (i,j)-th 
        diffraction order

        Parameters
        ----------

        i, j : int
               Diffraction order

        Returns
        -------

        efficiency : MuellerMatrix
                     Diffraction efficiency, as a Mueller matrix, at order (i,j)

        """
        return MuellerMatrix(
            SCATPY.CrossRCWDiffractionEfficiency(self.handle, i, j))

    def DiffractionAmplitude(self, i, j):
        """
        Returns the Jones matrix diffraction amplitude for the (i,j)-th 
        diffraction order

        Parameters
        ----------

        i, j : int
               Diffraction order

        Returns
        -------

        amplitude : numpy.array
                     Diffraction amplitude, as a Jones matrix, at order (i,j)

        """
        return np.array(
            SCATPY.CrossRCWDiffractionAmplitude(self.handle, i, j))

    def Direction(self,i,j):
        """
        Returns a 3D unit vector [x,y,z] in the direction of propagation of 
        the (i,j)-th diffraction order.

        Parameters
        ----------

        i, j : int
               Diffraction order
        
        Returns
        -------

        direction : list of float
                    Directional cosines of diffracted direction

        """
        return SCATPY.CrossRCWDirection(self.handle, i, j)




import numpy as np
from pySCATMECH.model import *
from pySCATMECH.mueller import *
import SCATPY 
import os

class OpticalFunction():
    """
    Class for handling optical functions.

    Parameters
    ----------
    name: string, complex, float, or function returning complex or float
        The complex index of refraction. If a string, it can be a 
        "(n,k)" pair or the name of a file containing a table of 
        wavelengths, n, and k.  

    """
    def __init__(self, name, iter=None):
        # if a function, store the function
        if callable(name):
            self.function = name
            # create a temporary file
            from random import randint
            import string
            # Make a 12 character random name starting with a letter
            chars = string.ascii_letters
            self.filename = chars[randint(0,len(chars)-1)]
            # Numbers can follow...
            chars = chars + string.digits
            for i in range(11):
                self.filename += chars[randint(0,len(chars)-1)]
            self.filename += ".tmp"
            try:
                with open(self.filename,"w") as file:
                    for L in iter:
                        nk = self.function(L)
                        file.write("%.16g\t%.16g\t%.16g\n" %
                                   (L,nk.real,nk.imag))
            except Exception as e:
                # if there was an exception, delete the file
                os.remove(self.filename)
                del self.filename
                raise
            self.name = self.filename
        else: 
            self.name = name
            
    def __del__(self):
        if hasattr(self,'filename'):
            try: 
                os.remove(self.filename)
            except:
                print("Problem removing",self.filename)
               

    def __call__(self, wavelength):
        """
        Evaluate the optical constants n and k at wavelength.
        
        Parameters
        ----------

        wavelength: float 
                    The wavelength
        Returns
        -------
                    The complex optical constant
        """
        if hasattr(self,"function"):
            return self.function(wavelength)
        if type(self.name) is str:
            return SCATPY.GetOpticalConstant(wavelength,self.name)
        return self.name
            
    def nkstr(self,wavelength=1):
        """
        Evaluate the optical constants n and k at wavelength and returns a 
        string that can be used by SCATMECH::dielectric_function. 

        Parameters
        ----------
        
        wavelength: float
                    The wavelength
        Returns
        -------
            : str
            The complex optical constant as a string "(n,k)" or the name of
            a file.
        """        
        if isinstance(self.name,str):
            return self.name
        nk = self(wavelength)
        return str((nk.real,nk.imag)).replace(" ","")
    
    def __str__(self):
        return str(self.name)
    
    def writeToFile(self,filename,wavelengths):
        """
        Write the optical function to a file that can be used by SCATMECH 
        routines as a SCATMECH::dielectric_function.  

        Parameters
        ----------

        filename: str
                  The name of the file
        wavelengths: iterable of float
                  The wavelengths that will be written.

        Returns
        -------
            str
            The filename, if successful, 
            `(n,k)` at the first wavelength if not successful
        """
        try:
            with open(filename,'w') as file:
                for wavelength in wavelengths:
                    nk = complex(self(wavelength))
                    file.write(str(wavelength) + "\t" +
                               str(nk.real)+ "\t" + str(nk.imag) + "\n")
            return filename
        except Exception as e:
            print(e)
            return self.nkstr(wavelengths[0])
        
class Film():
    """
    Class for handling an optical film. 

    Film stores an OpticalFunction and a thickness. One feature is that 
    multiplication by a scalar returns a Film with its thickness multiplied 
    by that scalar.

    Parameters
    ----------
    material: OpticalFunction
              The optical function for the material

    thickness: float, optional
               The thickness of the film

    waves : float, optional
            The number of waves for the layer

    wavelength : float, optional
                 The wavelength the layer should be optimized for, if  
                 `waves` is set 

    angle : float, optional
            Vacuum angle in radians the layer should be optimized for, if 
            `waves` is set (default is 0)
    """
    def __init__(self, material, thickness=None, waves=None, wavelength=None, angle=0 ):

        if type(material) is OpticalFunction:
            self.material = material
        else:
            self.material = OpticalFunction(material)
            
        if thickness is not None:
            self.thickness = thickness
            return
        if waves is not None:
            n = self.material(wavelength)
            angle_inside = cmath.asin(cmath.sin(angle)/n)
            t = wavelength/n.real*waves/np.cos(angle_inside)
            self.thickness = t.real    

    def __mul__(self,a):
        """Returns a Film with a thickness multiplied by `a`"""
        return Film(self.material,thickness = self.thickness*a)
    
    def __rmul__(self,a):
        """Returns a Film with a thickness multiplied by `a`"""
        return Film(self.material,thickness = self.thickness*a)

    def __str__(self):
        return str((self.material,self.thickness))
   
    
class FilmStack(Model):
    """A class for handling stacks of films"""
 
    def __init__(self,films=[]):
        """
        Parameters
        ----------

        films : list of Film
                A list of films. films[0] is closest to the substrate.

        """
        self.newer = SCATPY.Get_StackModel
        self.freer = SCATPY.Free_StackModel
        self.handle = SCATPY.Get_StackModel("Stack_StackModel")
        try: 
            self.films = []
            for f in films:
                self.films.append(f)
        except TypeError:
            self.films = [films]
        
    def grow(self,film):
        """
        Grow a film on the stack.

        Parameters
        ----------

        film : Film
               The film to be grown
        """
        self.films.append(film)
        
    def clean(self):
        """
        Clear the stack of all films.
        """
        self.films = []
        
    def getStackCommand(self,wavelength):
        """
        Get the list of films suitable for a SCATMECH::Stack_StackModel

        
        Parameters
        ----------

        wavelength : float
                     The wavelength to evaluate the complex index at.

        Returns
        -------
        
        command : str
                  List of films suitable for a SCATMECH::Stack_StackModel
        """
        self.command = ""
        for f in self.films:
            materialstring = str((f.material.nkstr(wavelength)).replace(" ",""))
            thicknessstring = str(f.thickness)
            self.command = (self.command + materialstring +
                            " " + thicknessstring + " ")
        return self.command

    def getModelDict(self):
        """
        Return a Model parameter dictionary. Used primarily internally so that
        a FilmStack can be passed as a parameter to a Model for a stack.
        """
        dictionary = {None: "Stack_StackModel"}
        dictionary["stack"] = self.getStackCommand(1)
        return dictionary
    
    def __str__(self):
        dictionary = self.getModelDict()
        return str(dictionary)
        
    def reflectionCoefficient(self, theta, wavelength, no, nt,type="12"):
        """
        Return the reflection coefficent.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        coeff : list of list of complex
                The Jones matrix reflection coefficient, which is the
                linear relationship between the electric field amplitudes.
                coeff[0][0] is the s-polarized reflection coefficient.
                coeff[1][1] is the p-polarized reflection coefficient.
                coeff[1][0] = coeff[0][1] = 0
        """
        self.setParameter("stack",self.getStackCommand(wavelength))
        return SCATPY.ReflectionCoefficient(
                             self.handle, theta, wavelength, 
                             no.nkstr(), 
                             nt.nkstr(),
                             type)
    
    def transmissionCoefficient(self, theta, wavelength, no, nt,type="12"):
        """
        Return the transmission coefficient.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        coeff : list of list of complex
                The Jones matrix transmission coefficient, which is the
                linear relationship between the electric field amplitudes.
                coeff[0][0] is the s-polarized transmission coefficient.
                coeff[1][1] is the p-polarized transmission coefficient.
                coeff[1][0] = coeff[0][1] = 0
        """
        self.setParameter("stack",self.getStackCommand(wavelength))
        return SCATPY.TransmissionCoefficient(
                             self.handle, theta, wavelength, 
                             no.nkstr(), 
                             nt.nkstr(),
                             type)

    def R(self, theta, wavelength, no, nt, type="12"):
        """
        Return the reflectance.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        refl : MuellerMatrix
               The Mueller matrix reflectance, which is the
               linear relationship between radiant fluxes.
        """
        return JonesMueller(self.reflectionCoefficient(theta,
                                                       wavelength,
                                                       no, nt,type))

    def T(self, theta, wavelength, no, nt, type="12"):
        """
        Return the transmittance.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        refl : MuellerMatrix
               The Mueller matrix transmittance, which is the
               linear relationship between radiant fluxes.
        """
        No = no(wavelength)
        Nt = nt(wavelength)
        thetat = np.arcsin(complex(np.sin(theta)*No/Nt))
        factor = abs(Nt*np.cos(thetat)/np.abs(np.cos(theta))/No)
        return JonesMueller(self.transmissionCoefficient(
                            theta, wavelength, no, nt,type))*factor
    
    def Rs(self,theta,wavelength,no,nt,type="12"):
        """
        Return the reflectance for s-polarized radiation.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        refl : float
               The reflectance for s-polarization, which is the
               linear relationship between radiant fluxes.
        """
        return (self.R(theta,wavelength,no,nt,type)*StokesVector([1,1,0,0]))[0]

    def Rp(self,theta,wavelength,no,nt,type="12"):
        """
        Return the reflectance for p-polarized radiation.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        refl : float
               The reflectance for p-polarization, which is the
               linear relationship between radiant fluxes.
        """
        return (self.R(theta,wavelength,no,nt,type)*StokesVector([1,-1,0,0]))[0]

    def Ts(self,theta,wavelength,no,nt,type="12"):
        """
        Return the transmittance for s-polarized radiation.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is `12`)

        Returns
        -------

        refl : float
               The transmittance for s-polarization, which is the
               linear relationship between radiant fluxes.
        """
        return (self.T(theta,wavelength,no,nt,type)*StokesVector([1,1,0,0]))[0]

    def Tp(self,theta,wavelength,no,nt,type="12"):
        """
        Return the transmittance for p-polarized radiation.

        Parameters
        ----------

        theta : float
                Incident angle in radians

        wavelength : float
                     Wavelength in vacuum

        no : OpticalFunction
             Optical constants for the incident medium

        nt : OpticalFunction
             Optical constants for the transmitted medium

        type : str, optional
               If `12`, then the incident medium is the medium above the 
               substrate and theta is the incident angle, evaluated as if 
               the radiation were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `21`, then the incident medium is the substrate and theta 
               is the incident angle evaluated is evaluated as if the radiation
               were in vacuum. (The component of the wavevector
               parallel to the surface is 2*pi*sin(theta)/lambda)
               If `12i`, then the incident medium is the medium above the 
               substrate and theta is the incident angle in that medium.
               (The component of the wavevector parallel to the surface is 
               2*pi*sin(theta)/no/lambda)
               If `21i`, then the incident medium is the substrate and theta 
               is the incident angle in the substrate. (The component of the 
               wavevector parallel to the surface is 2*pi*sin(theta)/n0/lambda)
               (Default is '12')

        Returns
        -------

        refl : float
               The transmittance for p-polarization, which is the
               linear relationship between radiant fluxes.
        """
        return (self.T(theta,wavelength,no,nt,type)*StokesVector([1,-1,0,0]))[0]





from setuptools import setup, Extension
#from distutils.core import setup, Extension
import platform

with open("README.md", "r") as fh:
    long_description = fh.read()

print("platform.system() = ",platform.system())

if platform.system() == 'Windows':
    extra_compile_args = []
    extra_link_args =[]
if platform.system() == 'Linux':
    extra_compile_args = ['-Wno-deprecated-declarations']
    extra_link_args =[]
elif 'CYGWIN' in platform.system():
    extra_compile_args = ['-Wno-deprecated-declarations']
    extra_link_args =[]

scatmech_source = ['scatmech/allrough.cpp',
                   'scatmech/crossgrating2.cpp',
                   'scatmech/fresnel.cpp',
                   'scatmech/models.cpp',
                   'scatmech/rcw.cpp',
                   'scatmech/scatmatrix.cpp',
                   'scatmech/two_source.cpp',
                   'scatmech/askuser.cpp',
                   'scatmech/crossrcw.cpp',
                   'scatmech/gcross.cpp',
                   'scatmech/mueller.cpp',
                   'scatmech/reflectance.cpp',
                   'scatmech/scattabl.cpp',
                   'scatmech/twoface.cpp',
                   'scatmech/axifree.cpp',
                   'scatmech/crough.cpp',
                   'scatmech/grating.cpp',
                   'scatmech/nsphere.cpp',
                   'scatmech/reg_brdf.cpp',
                   'scatmech/sizedistribution.cpp',
                   'scatmech/urough.cpp',
                   'scatmech/axipart1.cpp',
                   'scatmech/dielfunc.cpp',
                   'scatmech/inherit.cpp',
                   'scatmech/oasphere.cpp',
                   'scatmech/reg_facet.cpp',
                   'scatmech/sphdfct.cpp',
                   'scatmech/vector3d.cpp',
                   'scatmech/axipart2.cpp',
                   'scatmech/diffuse.cpp',
                   'scatmech/instrument.cpp',
                   'scatmech/onelayer.cpp',
                   'scatmech/reg_instrument.cpp',
                   'scatmech/sphprt.cpp',
                   'scatmech/zernike.cpp',
                   'scatmech/axisym.cpp',
                   'scatmech/facet.cpp',
                   'scatmech/jmatrix.cpp',
                   'scatmech/phasefunction.cpp',
                   'scatmech/reg_lambert.cpp',
                   'scatmech/sphrscat.cpp',
                   'scatmech/zernikeexpansion.cpp',
                   'scatmech/bobvlieg1.cpp',
                   'scatmech/fft.cpp',
                   'scatmech/jvector.cpp',
                   'scatmech/polydisperse.cpp',
                   'scatmech/reg_local.cpp',
                   'scatmech/stokes.cpp',
                   'scatmech/bobvlieg2.cpp',
                   'scatmech/filmtran.cpp',
                   'scatmech/lambert.cpp',
                   'scatmech/random.cpp',
                   'scatmech/reg_rough.cpp',
                   'scatmech/subbobvlieg.cpp',
                   'scatmech/bobvlieg3.cpp',
                   'scatmech/finiteaperture.cpp',
                   'scatmech/local.cpp',
                   'scatmech/raygscat.cpp',
                   'scatmech/reg_sphrscat.cpp',
                   'scatmech/subsphere.cpp',
                   'scatmech/brdf.cpp',
                   'scatmech/firstdiffuse.cpp',
                   'scatmech/matrixmath.cpp',
                   'scatmech/rayinst.cpp',
                   'scatmech/rough.cpp',
                   'scatmech/tmatrix.cpp',
                   'scatmech/coatedmie.cpp',
                   'scatmech/flake.cpp',
                   'scatmech/matrixmath2.cpp',
                   'scatmech/rayscat.cpp',
                   'scatmech/roughnes.cpp',
                   'scatmech/torrspar.cpp',
                   'scatmech/crossgrating.cpp',
                   'scatmech/focussedbeam.cpp',
                   'scatmech/miescat.cpp',
                   'scatmech/raystack.cpp',
                   'scatmech/scateval.cpp',
                   'scatmech/transmit.cpp']

scatmech_headers  = ['scatmech/allrough.h',
                     'scatmech/crossrcw.h',
                     'scatmech/flake.h',
                     'scatmech/matrix3d.h',
                     'scatmech/polydisperse.h',
                     'scatmech/roughnes.h',
                     'scatmech/subbobvlieg.h',
                     'scatmech/zernike.h',
                     'scatmech/askuser.h',
                     'scatmech/crough.h',
                     'scatmech/focussedbeam.h',
                     'scatmech/matrixmath.h',
                     'scatmech/random.h',
                     'scatmech/scateval.h',
                     'scatmech/subsphere.h',
                     'scatmech/zernikeexpansion.h',
                     'scatmech/axifree.h',
                     'scatmech/dielfunc.h',
                     'scatmech/fresnel.h',
                     'scatmech/miescat.h',
                     'scatmech/raygscat.h',
                     'scatmech/scatmatrix.h',
                     'scatmech/tmatrix.h',
                     'scatmech/axipart.h',
                     'scatmech/diffuse.h',
                     'scatmech/gcross.h',
                     'scatmech/mueller.h',
                     'scatmech/rayinst.h',
                     'scatmech/scatmech.h',
                     'scatmech/torrspar.h',
                     'scatmech/bobvlieg.h',
                     'scatmech/facet.h',
                     'scatmech/grating.h',
                     'scatmech/nsphere.h',
                     'scatmech/rayscat.h',
                     'scatmech/scattabl.h',
                     'scatmech/transmit.h',
                     'scatmech/brdf.h',
                     'scatmech/fft.h',
                     'scatmech/inherit.h',
                     'scatmech/oasphere.h',
                     'scatmech/raystack.h',
                     'scatmech/sizedistribution.h',
                     'scatmech/two_source.h',
                     'scatmech/coatedmie.h',
                     'scatmech/filmtran.h',
                     'scatmech/instrument.h',
                     'scatmech/onelayer.h',
                     'scatmech/rcw.h',
                     'scatmech/sphdfct.h',
                     'scatmech/twoface.h',
                     'scatmech/crossgrating.h',
                     'scatmech/finiteaperture.h',
                     'scatmech/lambert.h',
                     'scatmech/optconst.h',
                     'scatmech/reflectance.h',
                     'scatmech/sphprt.h',
                     'scatmech/urough.h',
                     'scatmech/crossgrating2.h',
                     'scatmech/firstdiffuse.h',
                     'scatmech/local.h',
                     'scatmech/phasefunction.h',
                     'scatmech/rough.h',
                     'scatmech/sphrscat.h',
                     'scatmech/vector3d.h']
                     

module1 = Extension('SCATPY',
                    include_dirs = ['scatmech'],
                    library_dirs = [],
                    libraries = [],
                    sources = ['SCATPYmodule.cpp'] + scatmech_source,
                    depends = scatmech_headers,
                    extra_compile_args = extra_compile_args,
                    extra_link_args =extra_link_args)

#setuptools.setup(
setup(
    name="pySCATMECH",
    version="0.1.8",  # NOTE: If you change this, change it in index.rst AND conf.py too!
    author="Thomas A. Germer",
    author_email="thomas.germer@nist.gov",
    description="A Python interface to the SCATMECH library",
    long_description=long_description,
    #long_description_content_type="text/markdown",
    url="https://github.com/thomas-germer/pyscatmech",
    packages=["pySCATMECH"],
    ext_modules=[module1],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: Public Domain",
        "Topic :: Scientific/Engineering :: Physics",
        "Operating System :: Microsoft :: Windows"
    ],
    python_requires='>=3.6'
)



__all__ = ['brdf',
           'local',
           'rcw',
           'crossrcw',
           'integrate',
           'model',
           'mueller',
           'scatterer',
           'fresnel'
           ]
